#ifndef CAMERA_3D_H__
#define CAMERA_3D_H__

#include "Frame3D.h"
#include <GLM/glm.hpp>

// A camera is a special case of a coordinate frame.
// We control this camera by the mouse. It can zoom, pan, and rotate in a
// similar way as the cameras used in Blender and Vicon Blade.
// It is very important to set the focus point correctly via the forward vector
// when constructing the camera, or via the center point when calling the LookAt
// function.
class Camera3D : public Frame3D {
public:
    Camera3D();
    virtual ~Camera3D();
    // See Frame3D.h.
    // Important: the length of the forward vector is now also used to
    // initialize the focus distance. See focusDistance_.
    Camera3D(const glm::vec3& origin,
             const glm::vec3& up,
             const glm::vec3& forward);

    // Sets the camera at (eyeX, eyeY, eyeZ), looking at
    // (centerX, centerY, centerZ), and the up vector is roughly in the
    // direction of (upX, upY, upZ).
    // This works similarly to the gluLookAt() function, except that the center
    // is now more important as it also specifies the focus point.
    void LookAt(float eyeX, float eyeY, float eyeZ,
                float centerX, float centerY, float centerZ,
                float upX, float upY, float upZ);

    // See Frame3D.h.
    void MoveForward(float d) override;

    // Rotate the camera around the axis that goes from the focus point to the
    // right of the screen (so seen from the camera, this axis is always
    // perfectly horizontal). Theta is given in degrees.
    void RotateX(float theta);
    // Rotate the camera around the axis that goes from the focus point in the
    // global Y direction. Theta is given in degrees.
    void RotateY(float theta);
protected:
    // The minimum focus distance. We don't want the camera to zoom in
    // infinitely.
    static const float kMinFocusDistance_;

    // The distance between the camera and the point in front of the camera
    // that will always be the center of rotation. This point moves along with
    // the camera and is always seen in the center of the screen.
    // This distance is set to some fixed value when we first place the camera
    // into the scene and it changes every time we zoom in/out.
    float focusDistance_;
};

#endif    // CAMERA_3D_H__