#ifndef EFFECTOR_H__
#define EFFECTOR_H__

#include <string>
#include "Eigen/Dense"
#include "Joint.h"

// An effector is a point on an IK tree that follows its target.
class Effector {
public:
    // See Joint.h.
    EIGEN_MAKE_ALIGNED_OPERATOR_NEW

    DUMMY_DEFAULT_CONSTRUCTOR_AND_DESTRUCTOR(Effector)

    // Creates an effector on the link between the specified joint and its
    // parent. If ratio is 1, the effector is at the joint, if ratio is 0, the
    // effector is at the parent. If joint is the root, ratio doesn't matter.
    Effector(const std::string& name, const Joint* joint, float ratio);

    // Gets the nearest joint that controls the effector.
    const Joint* GetParent() const;
    const std::string GetName() const;

    // Gets the target's position in the global frame.
    Eigen::Vector3f GetTargetPosition() const;
    // Gets the position of the effector in its parent's frame.
    Eigen::Vector3f GetLocalPosition() const;
    // Gets the position of the effector in the global frame.
    Eigen::Vector3f GetGlobalPosition() const;

    // Computes the position of the effector in the global frame.
    void ComputeGlobalPosition();

    // Sets a new global position for the target.
    void UpdateTarget(float newX, float newY, float newZ);

private:
    // A unique name.
    std::string name_;
    // The position of the effector in its parent's frame.
    Eigen::Vector3f localPosition_;
    // The position of the effector in the global frame.
    Eigen::Vector3f globalPosition_;
    // The position of the target in the global frame.
    Eigen::Vector3f globalTargetPosition_;
    // The effector's parent. It's the nearest joint that controls the effector.
    const Joint* parent_;
};

#endif    // EFFECTOR_H__