#ifndef IK_TREE_H__
#define IK_TREE_H__

#include <memory>
#include <boost/iterator/transform_iterator.hpp>
#include "Common.h"
#include "Tree.h"
#include "Effector.h"

// An IK tree consists of a normal tree and a set of effectors defined on that
// tree.
class EffectorCollection
{
private:
    // We need the below to use Boost's transform_iterator to convert an
    // iterator to pair<string, Effector> to an iterator to Effector.
    typedef std::map<std::string, Effector> EffMap;
    typedef std::pair<const std::string, Effector> StrEffPair;
    typedef Effector& (*GetValFunc)(StrEffPair& pair);
    typedef const Effector& (*GetCValFunc)(StrEffPair const& pair);
    typedef EffMap::iterator EffMapIter;
    typedef EffMap::const_iterator EffMapCIter;

public:
    // (Const) Iterator to Effector for internal use.
    typedef boost::transform_iterator<GetValFunc, EffMapIter> EffIter;
    typedef boost::transform_iterator<GetCValFunc, EffMapCIter> EffCIter;

    DUMMY_DEFAULT_CONSTRUCTOR_AND_DESTRUCTOR(EffectorCollection)

    // Gets the number of effector.
    int GetEffectorCount() const;

    void AddEffector(const Tree* tree, const std::string& name,
                     const std::string& jointName, float ratio);
    // Sets a new global position for the target whose name is name.
    void UpdateTarget(const std::string& name,
                      float newX, float newY, float newZ);

    // Computes the global positions and orientations of all the joints and
    // effectors.
    void ComputeGlobalData();

    // Functions to get (const) iterators to the beginning/end of the effector
    // list.
    EffIter Begin();
    EffIter End();
    EffCIter CBegin() const;
    EffCIter CEnd() const;

protected:
    // We use this to quickly locate an effector based on its unique name.
    EffMap effectors_;
};

// (Const) Iterator to Effector for external use.
typedef EffectorCollection::EffIter EffectorIterator;
typedef EffectorCollection::EffCIter EffectorIteratorConst;

#endif    // IK_TREE_H__