#ifndef FRAME_3D_H__
#define FRAME_3D_H__

#define GLM_SWIZZLE_XYZW

#include <GLM/glm.hpp>

// A general coordinate frame, specified in terms of the global coordinate
// frame. It can be used to easily model and control a virtual camera.
class Frame3D {
public:
    // Constructs the default frame at (0, 0, 0), Y up, looking down the -Z
    // axis.
    Frame3D();
    virtual ~Frame3D();
    // See Set()
    Frame3D(const glm::vec3& origin,
            const glm::vec3& up,
            const glm::vec3& forward);

    // Sets the position and orientation of the frame.
    // The up parameter only roughly specifies the direction of the "real" up
    // vector. The forward parameter needs not be normalized.
    void Set(const glm::vec3& origin,
             const glm::vec3& up,
             const glm::vec3& forward);

    // See matLocalWorld_.
    void BuildLocalToWorldMatrix();
    // See matWorldLocal_.
    void BuildWorldToLocalMatrix();
    // See matLocalWorld_. The corresponding
    // BuildLocalToWorldMatrix() must be called before this.
    glm::mat4 GetLocalToWorldMatrix() const;
    // See matWorldLocal_. The corresponding
    // BuildWorldToLocalMatrix() must be called before this.
    glm::mat4 GetWorldToLocalMatrix() const;

    // Moves the origin forward d units.
    virtual void MoveForward(float d);
    // Moves the origin to the right d units.
    void MoveRight(float d);
    // Moves the origin up d units.
    void MoveUp(float d);
    // Rotates the frame theta (degrees) around v. v is a vector in world frame
    // rooted at p. If p = (0, 0, 0), this is a "usual" rotation.
    void RotateWorld(float theta, const glm::vec3& p, const glm::vec3& v);

protected:
    // The origin of the frame in terms of the global frame.
    glm::vec3 origin_;
    // The (unit) up vector in terms of the global frame.
    glm::vec3 up_;
    // The (unit) forward vector in terms of the global frame.
    glm::vec3 forward_;
    // The matrix that transforms a point from this frame to the global frame.
    glm::mat4 matLocalWorld_;
    // The matrix that transforms a point from the global frame to this frame.
    glm::mat4 matWorldLocal_;
};

#endif    // FRAME_3D_H__