#include <cmath>
#include "IKSolverJacobian.h"

using namespace std;
using namespace Eigen;

IKSolverJacobian::IKSolverJacobian(Tree* tree, EffectorCollection* effectors)
  : IKSolver(tree, effectors)
{
    tree->ComputeGlobalData();

    // populate the index map
    int index = 3; // the first 3 columns are reserved for the root's translation
    FOR_CONST(joint, tree) {
        indexMap_[joint->GetID()] = index;
        index += joint->GetDoF();
    }

    // count the number of rows and columns
    int rowCount = 3 * effectors->GetEffectorCount();
    int columnCount = 3;
    FOR_CONST(joint, tree) {
        columnCount += joint->GetDoF();
    }

    // resize the vectors/matrices
    J_.resize(rowCount, columnCount);
    J_.setZero();
    e_.resize(rowCount);
    eClamped_.resize(rowCount);
    DeltaTheta_.resize(columnCount);
    DMaxi_.resize(effectors->GetEffectorCount());
    DMaxi_.fill(HUGE_VAL);

    // set dMax_ to half the average bone length
    float average = 0.0f;
    FOR_CONST(joint, tree) {
        average += joint->GetLocalPosition().norm();
    }
    average /= tree->GetJointCount();
    DMax_ = average / 2.0f;
}

void IKSolverJacobian::ComputeE(const EffectorCollection* effectors) {
    int effId = 0;
    FOR_CONST(eff, effectors) {
        Vector3f ds = eff->GetTargetPosition() - eff->GetGlobalPosition();
        e_(effId * 3    ) = ds.x();
        e_(effId * 3 + 1) = ds.y();
        e_(effId * 3 + 2) = ds.z();

        // clamp the length to dMax(i)
        float length = ds.norm();
        if (length > DMaxi_(effId)) {
            ds *= DMaxi_(effId) / length;
        }

        eClamped_(effId * 3    ) = ds.x();
        eClamped_(effId * 3 + 1) = ds.y();
        eClamped_(effId * 3 + 2) = ds.z();

        effId++;
    }
}

void IKSolverJacobian::BuildJacobian(const Tree* tree, const EffectorCollection* effectors) {
    int effId = 0;
    FOR_CONST(eff, effectors) {
        auto ancestor = eff->GetParent();
        while (ancestor) {
            // if this is the root and it can move
            if (ancestor->GetID() == 0 && tree->IsRootMovable()) {
                // x
                J_(effId * 3    ,  0) = 1.0f;
                J_(effId * 3    ,  1) = 0.0f;
                J_(effId * 3    ,  2) = 0.0f;
                // y
                J_(effId * 3 + 1,  0) = 0.0f;
                J_(effId * 3 + 1,  1) = 1.0f;
                J_(effId * 3 + 1,  2) = 0.0f;
                // z
                J_(effId * 3 + 2,  0) = 0.0f;
                J_(effId * 3 + 2,  1) = 0.0f;
                J_(effId * 3 + 2,  2) = 1.0f;
            }

            int jointIdx = indexMap_[ancestor->GetID()];
            auto diff = ancestor->GetGlobalPosition() - eff->GetGlobalPosition();
            for (int i = 0; i < ancestor->GetDoF(); ++i) {
                auto cross = diff.cross(ancestor->GetGlobalEulerAxis(i));
                J_(effId * 3    , jointIdx) = cross.x();
                J_(effId * 3 + 1, jointIdx) = cross.y();
                J_(effId * 3 + 2, jointIdx) = cross.z();
                jointIdx++;
            }

            ancestor = ancestor->GetParent();
        }

        effId++;
    }
}

void IKSolverJacobian::UpdateThetas(Tree* tree, EffectorCollection* effectors) {
    FOR(joint, tree) {
        Vector3f dTheta = kVecZero_;
        int jointIndex = indexMap_[joint->GetID()];
        for (int i = 0; i < joint->GetDoF(); ++i) {
            dTheta(i) = DeltaTheta_(jointIndex);
            jointIndex++;
        }
        joint->AddToEulerAngles(dTheta(0), dTheta(1), dTheta(2));
    }

    tree->TranslateRoot(DeltaTheta_(0), DeltaTheta_(1), DeltaTheta_(2));
    tree->ComputeGlobalData();
    effectors->ComputeGlobalData();
}

void IKSolverJacobian::ComputeDMaxi(const EffectorCollection* effectors) {
    int effId = 0;
    FOR_CONST(eff, effectors) {
        Vector3f newE = eff->GetTargetPosition() - eff->GetGlobalPosition();
        Vector3f oldE(e_(effId * 3), e_(effId * 3 + 1), e_(effId * 3 + 2));
        float dE = newE.norm() - oldE.norm();
        if (dE > 0.0f)
            DMaxi_(effId) = DMax_ + dE;
        else
            DMaxi_(effId) = DMax_;

        effId++;
    }
}

void IKSolverJacobian::Solve(Tree* tree, EffectorCollection* effectors,
                             int iterationCount) {
    for (int i = 0; i < iterationCount; ++i) {
        ComputeE(effectors);
        BuildJacobian(tree, effectors);
        ComputeDeltaTheta();
        UpdateThetas(tree, effectors);
        ComputeDMaxi(effectors);
    }
}