#include "Joint.h"

using namespace std;
using namespace Eigen;

Joint::Joint(int id, string name, float size, Joint* parent,
             const Vector3f& localPosition,
             const Vector3f& eulerAxis0, float eulerAngle0)
  : id_(id), name_(name), dof_(1), size_(size), parent_(parent),
    localPosition_(localPosition)
{
    localEulerAxes_[0] = eulerAxis0.normalized();

    eulerAngles_[0] = eulerAngle0 * kDegreeToRadian_;

    ComputeLocalQuaternion();
}

Joint::Joint(int id, string name, float size, Joint* parent,
            const Vector3f& localPosition,
            const Vector3f& eulerAxis0, float eulerAngle0,
            const Vector3f& eulerAxis1, float eulerAngle1)
  : id_(id), name_(name), dof_(2), size_(size), parent_(parent),
    localPosition_(localPosition)
{
    localEulerAxes_[0] = eulerAxis0.normalized();
    localEulerAxes_[1] = eulerAxis1.normalized();

    eulerAngles_[0] = eulerAngle0 * kDegreeToRadian_;
    eulerAngles_[1] = eulerAngle1 * kDegreeToRadian_;

    ComputeLocalQuaternion();
}

Joint::Joint(int id, string name, float size, Joint* parent,
             const Vector3f& localPosition,
             const Vector3f& eulerAxis0, float eulerAngle0,
             const Vector3f& eulerAxis1, float eulerAngle1,
             const Vector3f& eulerAxis2, float eulerAngle2)
  : id_(id), name_(name), dof_(3), size_(size), parent_(parent),
    localPosition_(localPosition)
{
    localEulerAxes_[0] = eulerAxis0.normalized();
    localEulerAxes_[1] = eulerAxis1.normalized();
    localEulerAxes_[2] = eulerAxis2.normalized();

    eulerAngles_[0] = eulerAngle0 * kDegreeToRadian_;
    eulerAngles_[1] = eulerAngle1 * kDegreeToRadian_;
    eulerAngles_[2] = eulerAngle2 * kDegreeToRadian_;

    ComputeLocalQuaternion();
}

int Joint::GetID() const { return id_; }

int Joint::GetDoF() const { return dof_; }

float Joint::GetSize() const { return size_; }

const Joint* Joint::GetParent() const { return parent_; }

Joint* Joint::GetParent() { return parent_; }

string Joint::GetName() const { return name_; }

Vector3f Joint::GetLocalPosition() const { return localPosition_; }

Vector3f Joint::GetLocalEulerAxis(int which) const {
    ASSERT(which >= 0 && which < dof_, "No such local Euler axis.");
    return localEulerAxes_[which];
}

float Joint::GetEulerAngle(int which) const {
    ASSERT(which >= 0 && which < dof_, "No such Euler angle.");
    return eulerAngles_[which];
}

Quaternionf Joint::GetLocalQuaternion() const { return localQuaternion_; }

Vector3f Joint::GetGlobalPosition() const { return globalPosition_; }

Quaternionf Joint::GetGlobalQuaternion() const { return globalQuaternion_; }

AngleAxisf Joint::GetGlobalAngleAxis() const {
    return AngleAxisf(globalQuaternion_);
}

AngleAxisf Joint::GetLocalAngleAxis() const {
    return AngleAxisf(localQuaternion_);
}

Vector3f Joint::GetGlobalEulerAxis(int which) const {
    ASSERT(which >= 0 && which < dof_, "No such global Euler axis.");
    return globalEulerAxes_[which];
}

void Joint::ComputeGlobalData() {
    ComputeGlobalQuaternion();
    ComputeGlobalPosition();
    ComputeGlobalEulerAxes();
}

void Joint::AddToEulerAngles(float dTheta0, float dTheta1, float dTheta2) {
    eulerAngles_[0] += dTheta0;
    eulerAngles_[1] += dTheta1;
    eulerAngles_[2] += dTheta2;

    ComputeLocalQuaternion();
}

void Joint::ComputeLocalQuaternion() {
    AngleAxisf aa(eulerAngles_[0], localEulerAxes_[0]);
    if (dof_ > 1)
        aa = aa * AngleAxisf(eulerAngles_[1], localEulerAxes_[1]);
    if (dof_ > 2)
        aa = aa * AngleAxisf(eulerAngles_[2], localEulerAxes_[2]);
    localQuaternion_ = aa;
}

void Joint::ComputeGlobalPosition() {
    if (parent_ != 0) { // if this joint is not the root
        globalPosition_ =
            parent_->globalQuaternion_._transformVector(localPosition_) +
            parent_->globalPosition_;
    } else { // if this joint is the root
        globalPosition_ = localPosition_;
    }
}

void Joint::ComputeGlobalQuaternion() {
    if (parent_ != 0) // if this joint is not the root
        globalQuaternion_ = parent_->globalQuaternion_ * localQuaternion_;
    else // if this joint is the root
        globalQuaternion_ = localQuaternion_;
}

void Joint::ComputeGlobalEulerAxes() {
    globalEulerAxes_[0] = localEulerAxes_[0];

    Quaternionf q1(AngleAxisf(eulerAngles_[0], localEulerAxes_[0]));
    globalEulerAxes_[1] = q1._transformVector(localEulerAxes_[1]);

    Quaternionf q2 = AngleAxisf(eulerAngles_[0], localEulerAxes_[0]) *
                     AngleAxisf(eulerAngles_[1], localEulerAxes_[1]);
    globalEulerAxes_[2] = q2._transformVector(localEulerAxes_[2]);

    if (parent_ != 0) { // if this is not the root
        Quaternionf q = parent_->GetGlobalQuaternion();
        globalEulerAxes_[0] = q._transformVector(globalEulerAxes_[0]);
        globalEulerAxes_[1] = q._transformVector(globalEulerAxes_[1]);
        globalEulerAxes_[2] = q._transformVector(globalEulerAxes_[2]);
    }
}

void Joint::AddChild(JointUPtr child) { children_.push_back(move(child)); }

void Joint::AddToLocalPosition(float dX, float dY, float dZ) {
    localPosition_ += Vector3f(dX, dY, dZ);
}