#ifndef JOINT_H__
#define JOINT_H__

#include <array>
#include <string>
#include <vector>
#include <Eigen/Dense>
#include "Common.h"

// A joint is the location at which two or more bones contact.
// We get from the coordinate frame of a joint's parent to its coordinate frame
// by first translating the parent's frame using the offset (local position)
// vector, then rotate the frame using (up to) 3 Euler angles around three
// particular axes (we call them Euler axes here). The Euler axes are given in
// terms of the parent's frame.
class Joint
{
private:
    typedef std::array<Eigen::Vector3f, 3> Array3Vector3f;
    typedef std::array<float, 3> Array3f;
    typedef std::unique_ptr<Joint> JointUPtr;
    typedef std::vector<JointUPtr> JointUPtrList;

public:
    // We need this because this class uses some fixed-size Eigen's structures
    // (such as Vector3f) as members, and those require a special "new" operator
    // to align data properly for SIMD to work. If we don't use SIMD (e.g. if we
    // prefer the program to be slower), this macro can be safely removed.
    EIGEN_MAKE_ALIGNED_OPERATOR_NEW

    DUMMY_DEFAULT_CONSTRUCTOR_AND_DESTRUCTOR(Joint)

    // Constructs a hinge joint (1 dof).
    Joint(int id, std::string name, float size, Joint* parent,
          const Eigen::Vector3f& localPosition,
          const Eigen::Vector3f& eulerAxis0, float eulerAngle0);
    // Constructs a saddle joint (2 dof).
    Joint(int id, std::string name, float size, Joint* parent,
          const Eigen::Vector3f& localPosition,
          const Eigen::Vector3f& eulerAxis0, float eulerAngle0,
          const Eigen::Vector3f& eulerAxis1, float eulerAngle1);
    // Construct a ball joint (3 dof).
    Joint(int id, std::string name, float size, Joint* parent,
          const Eigen::Vector3f& localPosition,
          const Eigen::Vector3f& eulerAxis0, float eulerAngle0,
          const Eigen::Vector3f& eulerAxis1, float eulerAngle1,
          const Eigen::Vector3f& eulerAxis2, float eulerAngle2);

    // Gets the id.
    int GetID() const;
    // Gets the degree of freedom (1, 2, or 3).
    int GetDoF() const;
    // Gets the size.
    float GetSize() const;
    // Gets a pointer to the (const) parent.
    const Joint* GetParent() const;
    // Gets a pointer to the parent.
    Joint* GetParent();
    std::string GetName() const;

    // Gets the local position.
    Eigen::Vector3f GetLocalPosition() const;
    // Gets the (first/second/third) local rotation axis.
    Eigen::Vector3f GetLocalEulerAxis(int which) const;
    // Gets the (first/second/third) Euler angles, in radian.
    float GetEulerAngle(int which) const;
    // Gets the local quaternion.
    Eigen::Quaternionf GetLocalQuaternion() const;

    // The functions below should only be called after calling the
    // ComputeGlobalData() function and before any changes are made to the joint
    // using AddToEulerAngles() or AddToLocalPosition().

    // Gets the joint's global position.
    Eigen::Vector3f GetGlobalPosition() const;
    // Gets the joint's global quaternion.
    Eigen::Quaternionf GetGlobalQuaternion() const;
    // Gets the joint's global orientation in terms of a rotation around a fixed
    // axis.
    Eigen::AngleAxisf GetGlobalAngleAxis() const;
    Eigen::AngleAxisf GetLocalAngleAxis() const;

    // Gets the joint's (first/second/third) global rotation axis.
    Eigen::Vector3f GetGlobalEulerAxis(int which) const;

    // Computes the global position and orientation. We don't compute this data
    // automatically since it's easier to be lazy...
    void ComputeGlobalData();

    // Adds dThetai to the ith Euler angle.
    void AddToEulerAngles(float dTheta1,
                          float dTheta2 = 0.0f,
                          float dTheta3 = 0.0f);

protected:
    friend class Tree;

    // Computes the local quaternion that rotates the parent's frame to this
    // joint's frame.
    void ComputeLocalQuaternion();
    // Computes the global position.
    void ComputeGlobalPosition();
    // Computes the global quaternion that rotates the global frame to this
    // joint's frame.
    void ComputeGlobalQuaternion();
    // Computes the Euler axes in terms of the global frame.
    void ComputeGlobalEulerAxes();

    // Adds a child.
    void AddChild(JointUPtr child);
    // Adds dX, dY, dZ to the local position. This only makes sense if the joint
    // is the root, in which case this means translating the root by
    // (dX, dY, dZ)
    void AddToLocalPosition(float dX, float dY, float dZ);

    // A unique id, starting from 0.
    int id_;
    // A unique name.
    std::string name_;
    // The degree of freedom (1, 2, or 3).
    int dof_;
    // The size, used in rendering.
    float size_;
    // The parent.
    Joint* parent_;
    // A list of children.
    JointUPtrList children_;

    // The position of the joint in the parent's coordinate frame.
    Eigen::Vector3f localPosition_;
    // Three Euler rotation axes, defined in terms of the parent's coordinate
    // frame.
    Array3Vector3f localEulerAxes_;
    // Three Euler angles, in radians.
    Array3f eulerAngles_;

    // If we rotate the parent's coordinate frame using this quaternion, we get
    // the current joint's coordinate frame (ignoring the translation part).
    Eigen::Quaternionf localQuaternion_;

    // The position of the joint in the fixed global coordinate frame.
    Eigen::Vector3f globalPosition_;
    // If we rotate the global coordinate frame using this quaternion, we get
    // the current joint's coordinate frame (ignoring the translation part).
    Eigen::Quaternionf globalQuaternion_;
    // Three Euler rotation axes, defined in terms of the global coordinate
    // frame.
    Array3Vector3f globalEulerAxes_;
};

#endif    // JOINT_H__