#include <iostream>
#include <GLEW/glew.h>
#include <GLFW/glfw.h>
#include <AntTweakBar/AntTweakBar.h>
#include <GLM/glm.hpp>
#include <GLM/ext.hpp>
#include "Tree.h"
#include "Program.h"
#include "Timer.h"
#include "AccumulativeTimer.h"
#include "FrameRate.h"
#include "Camera3D.h"
#include "Programs/AllPrograms.h"

using namespace std;
using namespace Eigen;
using namespace glm;

/* program_ is the test we want to run */
unique_ptr<Program> program_;

struct Mouse
{
    static enum MouseButtons { MOUSE_LEFT = 1, MOUSE_MIDDLE = 2, MOUSE_RIGHT = 4 };

    int activeButtons_;
    int mouseX_;
    int mouseY_;

    Mouse() :
        activeButtons_(0)
    {
    }
} mouse_;

/* The following variables are used to support controlling the camera with the mouse. */

struct Transform
{
    float rotateX_;
    float rotateY_;
    float translateX_;
    float translateY_;
    float translateZ_;

    Transform() :
        rotateX_(0.0f),
        rotateY_(0.0f),
        translateX_(0.0f),
        translateY_(0.0f),
        translateZ_(0.0f)
    {
    }

    void Reset()
    {
        rotateX_ = rotateY_ = 0.0f;
        translateX_ = translateY_ = translateZ_ = 0.0f;
    }
} transform_;

TwBar *programBar_;
Camera3D camera_;
Data data_;

struct ProgramState
{
    static enum States { PLAYING, PAUSED, STEPPING };

    int state_;
    double frameRate_;
    int programId_;

    ProgramState() :
        state_(PAUSED),
        frameRate_(0.0),
        programId_(2)
    {
    }

    void Reset()
    {
        state_ = PAUSED;
    }
} control_;

FrameRate frameRate_;

// do not use any fancy pointer here since the tweak bar manages its own object's memory
TwBar *tweakBar_;

const int kWindowWidth = 800;
const int kWindowHeight = 800;

const float zNear = 0.1f;
const float zFar = 1000.0f;
const float fov = 75.0f;

void CreateProgram(int id);

void RenderScene();
void ShutDown(int returnCode);

void InitGL();
void InitGLFW();
void InitTweakBar();

int GLFWCALL GLFWWindowClose();
void GLFWCALL GLFWChar(int character, int action);
void GLFWCALL GLFWKey(int key, int action);
void GLFWCALL GLFWMousePos(int x, int y);
void GLFWCALL GLFWMouseWheel(int pos);
void GLFWCALL GLFWMouseButton(int button, int action);
void GLFWCALL GLFWWindowResize(int width, int height);

/*
 * main
 */
int  main(int argc, char* argv[])
{
    // GLFW
    InitGLFW();

    // GLEW
    GLenum err = glewInit();
    if (err != GLEW_OK) {
        printf("ERROR: %s\n", glewGetErrorString(err));
        ShutDown(EXIT_FAILURE);
    }

    glutInit(&argc, argv);

    InitGL();

    // AntTweakBar
    InitTweakBar();

    data_.camera_ = &camera_;
    data_.bar_ = programBar_;

    CreateProgram(control_.programId_);

    while (true) {
        RenderScene();
        control_.frameRate_ = frameRate_.GetFrameRate();
    }

    ShutDown(EXIT_SUCCESS);
}

/*
 * Initializes GLFW and creates an OpenGL context
 */
void InitGLFW()
{
    if (glfwInit() != GL_TRUE) {
        printf("ERROR: glfwInit() failed.\n");
        ShutDown(EXIT_FAILURE);
    }

    glfwOpenWindowHint(GLFW_WINDOW_NO_RESIZE, GL_TRUE);
    if (glfwOpenWindow(kWindowWidth, kWindowHeight, 0, 0, 0, 0, 0, 0, GLFW_WINDOW) != GL_TRUE) {
        printf("ERROR: glfwOpenWindow() failed.\n");
        ShutDown(EXIT_FAILURE);
    }
    glfwSetWindowTitle("Animation");

    glfwEnable(GLFW_MOUSE_CURSOR);
    glfwEnable(GLFW_KEY_REPEAT);
    //glfwDisable(GLFW_AUTO_POLL_EVENTS);

    glfwSetKeyCallback(GLFWKey);
    glfwSetCharCallback(GLFWChar);
    glfwSetMousePosCallback(GLFWMousePos);
    glfwSetMouseButtonCallback(GLFWMouseButton);
    glfwSetWindowCloseCallback(GLFWWindowClose);
    glfwSetWindowSizeCallback(GLFWWindowResize);
}

/*
 * Terminates GLFW and the program
 */
void ShutDown(int returnCode)
{
    TwTerminate();
    glfwTerminate();
    exit(returnCode);
}

/*
 * Character key press callback for GLFW
 */
void GLFWCALL GLFWChar(int character, int action)
{
    if (!TwEventCharGLFW(character, action)) {
    }
}

/*
 * Key press callback for GLFW
 */
void GLFWCALL GLFWKey(int key, int action)
{
    if (!TwEventKeyGLFW(key, action)) {
    }

    if (key == GLFW_KEY_ESC && action == GLFW_PRESS)
        ShutDown(EXIT_SUCCESS);
}

/*
 * Mouse position callback for GLFW
 */
void GLFWCALL GLFWMousePos(int x, int y)
{
    // we let AntTweakBar handles the event first
    // if it doesn't, we handle it
    if (!TwEventMousePosGLFW(x, y)) {
        int dx = x - mouse_.mouseX_;
        int dy = y - mouse_.mouseY_;

        if (mouse_.activeButtons_ & Mouse::MOUSE_LEFT) {
            transform_.rotateX_ += dy;
            transform_.rotateY_ += dx;
            camera_.RotateX(0.5f * -dy);
            camera_.RotateY(0.5f * -dx);
        }

        if (mouse_.activeButtons_ & Mouse::MOUSE_MIDDLE) {
            transform_.translateZ_ += 0.1f * (dx - dy);
            camera_.MoveForward(0.1f * (dx - dy));
        }

        if (mouse_.activeButtons_ & Mouse::MOUSE_RIGHT) {
            transform_.translateX_ += dx * 0.05f;
            transform_.translateY_ += dy * 0.05f;
            camera_.MoveUp(0.01f * dy);
            camera_.MoveRight(0.01f * -dx);
        }

        mouse_.mouseX_ = x;
        mouse_.mouseY_ = y;
    }
}

/*
 * Mouse wheel callback for GLFW
 */
void GLFWCALL GLFWMouseWheel(int pos)
{
    if (!TwEventMouseWheelGLFW(pos)) {
    }
}

/*
 * Mouse button callback for GLFW
 */
void GLFWCALL GLFWMouseButton(int button, int action)
{
    // we let AntTweakBar handles the event first
    // if it doesn't, we handle it
    if (!TwEventMouseButtonGLFW(button, action)) {
        int b;

        switch (button) {
        case GLFW_MOUSE_BUTTON_LEFT:
            b = Mouse::MOUSE_LEFT;
            break;
        case GLFW_MOUSE_BUTTON_RIGHT:
            b = Mouse::MOUSE_RIGHT;
            break;
        case GLFW_MOUSE_BUTTON_MIDDLE:
            b = Mouse::MOUSE_MIDDLE;
            break;
        }

        if (action == GLFW_PRESS) {
            mouse_.activeButtons_ |= b;
        } else {
            mouse_.activeButtons_ &= ~b;
        }
    }
}

/*
 * Window close callback for GLFW
 */
int GLFWCALL GLFWWindowClose()
{
    ShutDown(EXIT_SUCCESS);
    return EXIT_SUCCESS;
}

/*
 * Window resize callback for GLFW
 */
void GLFWCALL GLFWWindowResize(int width, int height)
{
    glViewport(0, 0, width, height);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(fov, float(width) / float(height), zNear, zFar);
    glMatrixMode(GL_MODELVIEW);

    TwWindowSize(width, height);
}

/*
 * Initializes some OpenGL states
 */
void InitGL()
{
    glClearColor(0.5f, 0.5f, 0.5f, 1.0f);
    glColorMaterial (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_COLOR_MATERIAL);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glEnable(GL_NORMALIZE);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glShadeModel(GL_SMOOTH);
}

void TW_CALL Pause(void *data)
{
    if (control_.state_ == ProgramState::PLAYING) {
        control_.state_ = ProgramState::PAUSED;
        TwDefine("TweakBar/Pause label='Play'");
    } else if (control_.state_ == ProgramState::PAUSED) {
        control_.state_ = ProgramState::PLAYING;
        TwDefine("TweakBar/Pause label='Pause'");
    }
}

void TW_CALL Step(void *data)
{
    if (control_.state_ == ProgramState::PAUSED) {
        control_.state_ = ProgramState::STEPPING;
    }
}

void TW_CALL Reset(void *data)
{
    control_.Reset();
    TwDefine("TweakBar/Pause label='Play'");
    transform_.Reset();
    CreateProgram(control_.programId_);
}

void TW_CALL SetProgram(const void *value, void *data)
{
    int newProgramId = *(const int *) value;

    if (newProgramId != control_.programId_) {
        CreateProgram(newProgramId);
    }

    control_.programId_ = newProgramId;
}

void TW_CALL GetProgram(void *value, void *data)
{
    *(int *) value = control_.programId_;
}

void TW_CALL SetUpdateFrequency(const void *value, void *data)
{
    program_->SetUpdateFrequency(*(const int *) value);
}

void TW_CALL GetUpdateFrequency(void *value, void *data)
{
    *(int *) value = program_->GetUpdateFrequency();
}

void TW_CALL GetFloorEnable(void *value, void *data)
{
    *(bool *) value = program_->GetRenderer()->IsEnable(Renderer::FLOOR);
}

void TW_CALL SetFloorEnable(const void *value, void *data)
{
    bool currentValue = *(const bool *) value;
    if (currentValue) {
        program_->GetRenderer()->Enable(Renderer::FLOOR);
    } else {
        program_->GetRenderer()->Disable(Renderer::FLOOR);
    }
}

void ResetProgramBar()
{
    if (data_.bar_) {
        TwDeleteBar(data_.bar_);
    }

    data_.bar_ = TwNewBar("ProgramBar");
    TwDefine(" ProgramBar position='580 16' "
        " label='Program control' help='This bar explains some of the control buttons and variables' ");
}

void CreateProgram(int id)
{
    ResetProgramBar();

    switch (id) {
        case 0:
            program_.reset(new TwoBallJoints(&data_));
            break;
        case 1:
            program_.reset(new HandIK(&data_));
            break;
        case 2:
            program_.reset(new MultipleJointTypes(&data_));
            //const float lambda = 0.4;
            //static_cast<IKProgram *>(program_.get())->SetDLSLambda(&lambda, &data_);
            break;
        case 3:
            program_.reset(new ThreeHingeJoints(&data_));
            break;
        case 4:
            program_.reset(new EffectorOnBone(&data_));
            break;
        default:
            assert(false);
    }
}

void InitTweakBar()
{
    TwInit(TW_OPENGL, NULL);
    tweakBar_ = TwNewBar("TweakBar");
    TwDefine(" TweakBar label='Main control' help='This bar explains some of the control buttons and variables' ");

    TwAddVarRO(tweakBar_, "FrameRate", TW_TYPE_DOUBLE, &control_.frameRate_, " label='FPS' ");
    TwAddButton(tweakBar_, "Pause", Pause, NULL, " label='Play' ");
    TwAddButton(tweakBar_, "Step", Step, NULL, " label='Step' ");
    TwAddButton(tweakBar_, "Reset", Reset, NULL, " label='Reset' ");
    TwAddVarCB(tweakBar_, "UpdateStepsPerSecond", TW_TYPE_INT32, SetUpdateFrequency, GetUpdateFrequency, NULL,
        " min=15 max=240 step=15 label='Animation FPS' help='Number of animation update steps per second' ");
    TwAddVarCB(tweakBar_, "Show floor", TW_TYPE_BOOLCPP, SetFloorEnable, GetFloorEnable, NULL,
        " label='Show floor'");

    TwType programIdType;
    programIdType = TwDefineEnumFromString("ProgramIdType",
        "0. Two ball joints,"
        "1. Hand IK,"
        "2. Multiple Joint Types,"
        "3. Three Hinge Joints,"
        "4. Effectors on Bone"
    );

    TwAddVarCB(tweakBar_, "Program", programIdType, SetProgram, GetProgram, NULL, " label='Program' ");
}

void RenderScene()
{
    frameRate_.StartFrame();

    static AccumulativeTimer updateTimer_;
    updateTimer_.Start();

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

    glPushMatrix();

    program_->BeginFrame();

    camera_.BuildWorldToLocalMatrix();
    mat4 M = camera_.GetWorldToLocalMatrix();
    glMultMatrixf(value_ptr(M));

    program_->Draw();

    // do animation update if it's time to do so
    if (control_.state_ == ProgramState::PLAYING) {
        if (updateTimer_.GetElapsedTime() >= 1.0 / program_->GetUpdateFrequency()) {
            program_->Update(glfwGetTime());
            updateTimer_.Reset();
        }
    } else if (control_.state_ == ProgramState::STEPPING) {
        program_->Update(glfwGetTime());
        updateTimer_.Reset();
        control_.state_ = ProgramState::PAUSED;
    }

    glPopMatrix();

    TwDraw();

    glfwSwapBuffers();

    updateTimer_.Stop();

    frameRate_.EndFrame();
}