#include <fstream>
#include "Common.h"
#include "HandIK.h"
#include "Eigen/Dense"

using namespace std;
using namespace Eigen;
using namespace glm;

HandIK::HandIK(Data* data) :
    IKProgram(data)
{
    name_ = "Hand IK";
    description_ = "Test hand IK...";

    const char* skeFile = "../../Data/Motions/hand_skeleton_20_ball_out.txt";

    ReadSkeleton(skeFile, &skePos_, &skeRots_, &skeLinks_, &skeDofs_, &skeEffectors_);

    // build the tree
    ResetTree();
    const float jointSize = 0.02f; // no reason...
    int numJoints = skePos_.size();
    FOR_EACH(joint, 0, numJoints) {
        AddJoint(joint, skePos_, skeRots_, skeLinks_, skeDofs_);
    }
    tree_->EnableRootTranslation();
    // create effectors
    ResetEffectors();
    int numEffectors = skeEffectors_.size();
    FOR_EACH(effector, 0, numEffectors) {
        string effectorName = IntToString(effector);
        string jointName = IntToString(skeEffectors_[effector]);
        effectors_->AddEffector(tree_.get(), jointName, jointName, 1.0f);
    }

    //  read the effectors' motion
    ReadMotion("../../Data/Motions/hand_mocap_a5.ik.txt", &motion_);

    ResetSolver(IKSolver::DLS);
    ResetRenderer();

    errors_.resize(numEffectors);
    FOR_EACH(i, 0, errors_.size()) {
        errors_[i] = 0.0f;
    }
    //WriteTree("./Data/Motions/tree.txt");
}

void HandIK::AddJoint(int joint, const VecVec3& skePos, const VecVec3& skeRots,
                      const VecIntInt& skeLinks, const VecInt& skeDofs)
{
    string jointName = IntToString(joint);

    // if this joint is already added, do nothing
    if (tree_->Contain(jointName)) {
        return;
    }

    // find the parent
    int parent = -1;
    for (int j = 0; j < skeLinks.size(); ++j) {
        if (joint == skeLinks[j].second && joint != skeLinks[j].first) {
            parent = skeLinks[j].first;
            break;
        }
    }

    string parentName = IntToString(parent);
    if (parent != -1 && !tree_->Contain(parentName)) {
        AddJoint(parent, skePos, skeRots, skeLinks, skeDofs);
    }

    Vector3f pos(skePos[joint].x, skePos[joint].y, skePos[joint].z);
    vec3 rot = skeRots[joint] * kRadianToDegree_;
    int dof = skeDofs[joint]; // doesn't really matter, always 3
    if (dof == 3) {
        tree_->AddJoint(jointName, 0.02f, parentName, pos, kVecUnitX_, rot.x,
                        kVecUnitY_, rot.y, kVecUnitZ_, rot.z);
    } else if (dof == 1) {
        tree_->AddJoint(jointName, 0.02f, parentName, pos, kVecUnitX_, rot.x);
    } else if (dof == 2) {
        tree_->AddJoint(jointName, 0.02f, parentName, pos, kVecUnitX_, rot.x,
                        kVecUnitZ_, rot.z);
    }
}

void HandIK::Update(float time)
{
    static VecVecAA rots(motion_.size());
    static VecVec3 rootPos;

    if (step_ < motion_.size() - 1) {
        FOR_EACH(effector, effectors_->Begin(), effectors_->End()) {
            int id = StringToInt(effector->GetName());
            vec3 pos = motion_[step_][id];
            string name = effector->GetName();
            effectors_->UpdateTarget(name, pos.x, pos.y, pos.z);
        }

        solver_->Solve(tree_.get(), effectors_.get(), iterationCount_);

        FOR_EACH(effector, effectors_->Begin(), effectors_->End()) {
            float error = (effector->GetTargetPosition() - effector->GetGlobalPosition()).norm();
            int id = StringToInt(effector->GetName());
            errors_[id] += error;
        }

        Vector3f pos = tree_->GetRoot()->GetGlobalPosition();
        rootPos.push_back(vec3(pos.x(), pos.y(), pos.z()));

        rots[step_].resize(skeRots_.size());
        FOR_EACH(joint, tree_->Begin(), tree_->End()) {
            int id = StringToInt(joint->GetName());
            rots[step_][id] = joint->GetLocalAngleAxis();
        }

        ++step_;
    } else if (step_ == motion_.size() - 1) {
        /*WriteMotion("./Data/Motions/hand_motion_21_angle_axis.txt", rootPos, rots);

        float totalError = 0.0f;
        FOR_EACH(i, 0, errors_.size()) {
            totalError += errors_[i];
        }
        cout << "Error: " << (totalError / effectors_->GetEffectorCount()) / motion_.size() << endl;*/

        ++step_;
    }
}

void HandIK::ReadSkeleton(const char* skeFile, VecVec3* skePos, VecVec3* skeRots,
                          VecIntInt* skeLinks, VecInt* skeDofs, VecInt* skeEffectors)
{
    ifstream skeStream(skeFile);
    if (skeStream.fail()) {
        cerr << "ERROR: Fail to open " << skeFile << "." << endl;
        return;
    }

    int numJoints;
    skeStream >> numJoints;

    // joint positions and frames
    for (int i = 0; i < numJoints; i++) {
        vec3 pos;
        skeStream >> pos.x >> pos.y >> pos.z;
        skePos->push_back(pos);

        vec3 rot;
        skeStream >> rot.x >> rot.y >> rot.z;
        skeRots->push_back(rot);
    }

    // joint dofs
    for (int i = 0; i < numJoints; i++) {
        int dof;
        skeStream >> dof;
        skeDofs->push_back(dof);
    }

    // bones
    for (int i = 0; i < numJoints - 1; i++) {
        IntInt link;
        skeStream >> link.first >> link.second;
        skeLinks->push_back(link);
    }

    // effectors
    int numEffectors;
    skeStream >> numEffectors;
    for (int i = 0; i < numEffectors; i++) {
        int effectorNo;
        int effectorId;
        skeStream >> effectorNo >> effectorId;
        skeEffectors->push_back(effectorId);
    }

    skeStream.close();
}

void HandIK::ReadMotion(const char* motionFile, VecVecVec3* motion)
{
    ifstream myStream(motionFile);
    if (myStream.fail()) {
        cerr << "ERROR: Cannot open file " << motionFile << "." << endl;
        return;
    }

    string line;
    getline(myStream, line);
    stringstream ss(line);
    int numPoints = 0;
    string item;
    while(ss >> item) {
        numPoints++;
    }
    numPoints /= 3;

    while(getline(myStream, line)) {
        stringstream sss(line);
        int temp;
        sss >> temp; // the frame number, which we ignore

        VecVec3 frame;
        FOR_EACH(pointId, 0, numPoints) {
            vec3 point;
            sss >> point.x >> point.y >> point.z;
            frame.push_back(point);
        }
        motion->push_back(frame);
    }

    myStream.close();
}

// Normalizes an angle to the range [minValue, maxValue]
float HandIK::NormalizeAngle(float value, float minValue, float maxValue)
{
    float period = maxValue - minValue;
    assert(period > 0.1f); // we don't want minValue == maxValue...

    while (value < minValue) {
        value += period;
    }

    while (value > maxValue) {
        value -= period;
    }

    assert(minValue <= value && value <= maxValue);

    return value;
}

void HandIK::WriteMotion(const char* skeFile, const VecVec3& rootPos,
                         const VecVecAA& skeRots)
{
    ofstream skeStream(skeFile);
    if (skeStream.fail()) {
        cerr << "ERROR: Fail to open " << skeFile << "." << endl;
        return;
    }

    int numFrames = rootPos.size();
    int numJoints = skeRots[0].size();

    skeStream << numJoints << " " << numFrames << endl;

    FOR_EACH(frame, 0, numFrames) {
        skeStream << rootPos[frame].x << " " << rootPos[frame].y << " " << rootPos[frame].z << " " ;
        FOR_EACH(joint, 0, numJoints) {
            AngleAxisf aa = skeRots[frame][joint];
            float angle = aa.angle();
            NormalizeAngle(angle, -kPi_, kPi_);
            Vector3f axis = aa.axis();
            skeStream << angle << " " << axis.x() << " " << axis.y() << " " << axis.z() << " ";
        }
        skeStream << endl;
    }

    skeStream.close();
}

void HandIK::WriteTree(const char* fileName)
{
    struct Item
    {
        Vector3f xAxis;
        Vector3f yAxis;
        Vector3f zAxis;
        Vector3f pos;
    };

    vector<Item> allItems(20);

    ofstream myStream(fileName);
    if (myStream.fail()) {
        cerr << "ERROR: Cannot open file " << fileName << "." << endl;
        return;
    }

    tree_->ComputeGlobalData();

    FOR_EACH(joint, tree_->Begin(), tree_->End()) {
        Quaternionf q = joint->GetGlobalQuaternion();
        Item item;
        item.xAxis = q._transformVector(kVecUnitX_);
        item.yAxis = q._transformVector(kVecUnitY_);
        item.zAxis = q._transformVector(kVecUnitZ_);
        item.pos = joint->GetGlobalPosition();
        int jointId = StringToInt(joint->GetName());
        allItems[jointId] = item;
    }

    FOR_EACH(item, allItems.begin(), allItems.end()) {
        myStream << item->pos.x() << " " << item->pos.y() << " " << item->pos.z() << endl;
        myStream << item->xAxis.x() << " " << item->xAxis.y() << " " << item->xAxis.z() << endl;
        myStream << item->yAxis.x() << " " << item->yAxis.y() << " " << item->yAxis.z() << endl;
        myStream << item->zAxis.x() << " " << item->zAxis.y() << " " << item->zAxis.z() << endl;
        myStream << endl;
    }

    myStream.close();
}