#include <GLEW/glew.h>
#include <GLFW/glfw.h>
#include <Eigen/Dense>
#include <GLM/ext.hpp>
#include "Renderer.h"

using namespace std;
using namespace Eigen;
using namespace glm;

Renderer::Renderer() {}

Renderer::~Renderer() {
    glDeleteLists(shape_[SPHERE], 1);
    glDeleteLists(shape_[CYLINDER], 1);
    glDeleteTextures(1, &floorTexture_);
}

Renderer::Renderer(Camera3D *camera) : camera_(camera) {
    SetDefaultColor();
    SetDefaultEnable();
    SetUpLights();
    CreateShapes();
    CreateFloorTexture();
}

void Renderer::SetDefaultColor() {
    color_[JOINT] = vec4(0.12f, 0.56f, 1.00f, 1.00f);
    color_[BONE] = vec4(0.12f, 0.56f, 1.00f, 1.00f);
    color_[TARGET] = vec4(1.00f, 0.00f, 0.00f, 0.50f);
    color_[TARGET_LINE] = vec4(1.00f, 1.00f, 0.00f, 1.00f);
}

void Renderer::SetDefaultEnable() {
    isEnabled_[FLOOR      ] = false;
    isEnabled_[GLOBAL_AXES] = true ;
    isEnabled_[JOINT      ] = true ;
    isEnabled_[BONE       ] = true ;
    isEnabled_[TARGET     ] = true ;
    isEnabled_[TARGET_LINE] = true ;
    isEnabled_[LOCAL_FRAME] = true ;
}

void Renderer::CreateShapes() {
    shape_[SPHERE] = glGenLists(1);
    glNewList(shape_[SPHERE], GL_COMPILE);
    glutSolidSphere(1.0, 32, 32);
    glEndList();

    shape_[CYLINDER] = glGenLists(1);
    glNewList(shape_[CYLINDER], GL_COMPILE);
    glutSolidCylinder(1.0, 1.0, 32, 16);
    glEndList();
}

void Renderer::SetUpLights() {
    light_.ambient = vec4(0.2f, 0.2f, 0.2f, 1.0f);
    light_.diffuse = vec4(0.8f, 0.8f, 0.8f, 1.0f);
    light_.specular = vec4(1.0f, 1.0f, 1.0f, 1.0f);
    light_.position = vec4(1.0f, 1.0f, 1.0f, 1.0f);

    glLightfv(GL_LIGHT0, GL_AMBIENT, value_ptr(light_.ambient));
    glLightfv(GL_LIGHT0, GL_DIFFUSE, value_ptr(light_.diffuse));
    glLightfv(GL_LIGHT0, GL_SPECULAR, value_ptr(light_.specular));
    glLightfv(GL_LIGHT0, GL_POSITION, value_ptr(light_.position));
}

void Renderer::SetColor(Renderable which, const vec4& color) {
    color_[which] = color;
}

void Renderer::Enable(Renderable which) { isEnabled_[which] = true; }

void Renderer::Disable(Renderable which) { isEnabled_[which] = false; }

bool Renderer::IsEnable(Renderable which) const { return isEnabled_[which]; }

void Renderer::SetCameraLookAt(float eyeX, float eyeY, float eyeZ,
                               float centerX, float centerY, float centerZ,
                               float upX, float upY, float upZ)
{
    camera_->LookAt(eyeX, eyeY, eyeZ, centerX, centerY, centerZ, upX, upY, upZ);
}

void Renderer::DrawCylinder(Eigen::Vector3f axis, float radius) const {
    glPushMatrix();
    float length = axis.norm();
    axis.normalize();
    float angle = acos(axis.z());
    axis = axis.cross(Vector3f::UnitZ());
    glRotatef(-angle * kRadianToDegree_, axis.x(), axis.y(), axis.z());
    glTranslatef(0.0f, 0.0f, -length / 2.0f);
    glScalef(radius, radius, length);
    glCallList(shape_[CYLINDER]);
    glPopMatrix();
}

void Renderer::DrawTree(const Tree *tree) const {
    for (auto joint = tree->CBegin(); joint != tree->CEnd(); ++joint) {
        glPushMatrix(); // we are in the world frame

        auto position = joint->GetGlobalPosition();
        glTranslatef(position.x(), position.y(), position.z());

        glPushMatrix(); // now we have translated to the joint

        auto parent = joint->GetParent();
        if (parent) {
            AngleAxisf angleAxis = parent->GetGlobalAngleAxis();
            glRotatef(angleAxis.angle() * kRadianToDegree_, angleAxis.axis().x(), angleAxis.axis().y(), angleAxis.axis().z());
        }

        glPushMatrix(); // now we have rotated to the parent's local frame

        float size = joint->GetSize();
        glColor4fv(value_ptr(color_[JOINT]));
        if (joint->GetDoF() == 1) {
            Vector3f axis = joint->GetLocalEulerAxis(0);
            DrawCylinder(axis / axis.norm() * size * 4.0f, size);
        } else if (joint->GetDoF() == 2) {
            AngleAxisf angleAxis = joint->GetLocalAngleAxis();
            glRotatef(angleAxis.angle() * kRadianToDegree_, angleAxis.axis().x(), angleAxis.axis().y(), angleAxis.axis().z());
            Vector3f firstRotationAxis = joint->GetLocalEulerAxis(0);
            float firstRotationAngle = joint->GetEulerAngle(0);
            DrawCylinder(firstRotationAxis / firstRotationAxis.norm() * size * 4.0f, size);
            //glRotatef(firstRotationAngle * kRadianToDegree_, firstRotationAxis.x(), firstRotationAxis.y(), firstRotationAxis.z());
            Vector3f secondRotationAxis = joint->GetLocalEulerAxis(1);
            DrawCylinder(secondRotationAxis / secondRotationAxis.norm() * size * 4.0f, size);
        } else if (joint->GetDoF() == 3) {
            glScalef(size * 2.0f, size * 2.0f, size * 2.0f);
            glCallList(shape_[SPHERE]);
        }

        glPopMatrix(); // we are back to the parent's local frame

        // draw the local coordinate frame
        if (isEnabled_[LOCAL_FRAME]) {
            AngleAxisf angleAxis = joint->GetLocalAngleAxis();
            glRotatef(angleAxis.angle() * kRadianToDegree_, angleAxis.axis().x(), angleAxis.axis().y(), angleAxis.axis().z());

            DrawGlobalAxes(size * 2.0f);
        }

        glPopMatrix(); // we are back to the joint without any rotation

        // draws a cylinder from each joint to its parent
        if (parent) {
            auto parentPosition = parent->GetGlobalPosition();
            auto difference = parentPosition - position;
            glTranslatef(difference.x() / 2.0f, difference.y() / 2.0f, difference.z() / 2.0f);
            glColor4fv(value_ptr(color_[BONE]));
            DrawCylinder(difference, parent->GetSize());
        }

        glPopMatrix(); // we are back to the world frame
    }
}

void Renderer::DrawTargets(const EffectorCollection *ikTree, bool drawLines) const {
    for (auto effector = ikTree->CBegin(); effector != ikTree->CEnd(); ++effector) {
        auto position = effector->GetGlobalPosition();
        auto targetPosition = effector->GetTargetPosition();

        glPushMatrix();

        glTranslatef(targetPosition.x(), targetPosition.y(), targetPosition.z());
        glColor4fv(value_ptr(color_[TARGET]));
        float size = effector->GetParent()->GetSize();
        glScalef(size * 2.0f, size * 2.0f, size * 2.0f);
        glCallList(shape_[SPHERE]);

        glPopMatrix();

        if (drawLines) {
            glDisable(GL_LIGHTING);
            glDisable(GL_DEPTH_TEST);
            glColor4fv(value_ptr(color_[TARGET_LINE]));
            glBegin(GL_LINES);
            glVertex3f(position.x(), position.y(), position.z());
            glVertex3f(targetPosition.x(), targetPosition.y(), targetPosition.z());
            glEnd();
            glEnable(GL_LIGHTING);
            glEnable(GL_DEPTH_TEST);
        }
    }
}

void Renderer::DrawGlobalAxes(float length) const {
    if (!isEnabled_[GLOBAL_AXES])
        return;

    glDisable(GL_DEPTH_TEST);
    glDisable(GL_LIGHTING);

    glLineWidth(2.0f);
    glBegin(GL_LINES);
        glColor4f(1.0f, 0.0f, 0.0f, 0.5f);
        glVertex3f(0.0f, 0.0f, 0.0f);
        glVertex3f(length, 0.0f, 0.0f);
        glColor4f(0.0f, 1.0f, 0.0f, 0.5f);
        glVertex3f(0.0f, 0.0f, 0.0f);
        glVertex3f(0.0f, length, 0.0f);
        glColor4f(0.0f, 0.0f, 1.0f, 0.5f);
        glVertex3f(0.0f, 0.0f, 0.0f);
        glVertex3f(0.0f, 0.0f, length);
    glEnd();

    glLineWidth(1.0f);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
}

void Renderer::CreateFloorTexture() {
    glGenTextures(1, &floorTexture_);
    glBindTexture(GL_TEXTURE_2D, floorTexture_);
    glfwLoadTexture2D("../../Data/Textures/tile.tga", GLFW_BUILD_MIPMAPS_BIT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
}

void Renderer::DrawFloor(float size) const {
    if (!isEnabled_[FLOOR])
        return;

    glEnable(GL_TEXTURE_2D);

    glColor3f(0.8f, 0.8f, 0.8f);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glBindTexture(GL_TEXTURE_2D, floorTexture_);

    glBegin(GL_QUADS);
    glTexCoord2f(-size / 2.0f, -size / 2.0f);
    glVertex3f(-size, 0.0f, -size);
    glTexCoord2f(-size / 2.0f, size / 2.0f);
    glVertex3f(-size, 0.0f, size);
    glTexCoord2f(size / 2.0f, size / 2.0f);
    glVertex3f(size, 0.0f, size);
    glTexCoord2f(size / 2.0f, -size / 2.0f);
    glVertex3f(size, 0.0f, -size);
    glEnd();

    glDisable(GL_TEXTURE_2D);
}