// TODO: load from BVH?
// TODO: construct from global pos?

#include "Common.h"
#include "Tree.h"

using namespace boost;
using namespace std;
using namespace Eigen;

Tree::Tree() : isRootMovable_(false) {}

Tree::~Tree() {}

void Tree::AddJoint(const std::string& name, float size,
                    const std::string& parentName,
                    const Eigen::Vector3f& localPosition,
                    const Eigen::Vector3f& eulerAxis0, float eulerAngle0)
{
    if (jointList_.size() == 0) { // this is the root
        root_.reset(new Joint(jointList_.size(), name, size, nullptr,
            localPosition, eulerAxis0, eulerAngle0));
        AddJointToMapList(root_.get());
    } else { // this is not the root
        auto parent = GetJoint(parentName);
        ASSERT(parent != nullptr, "Parent does not exist.");
        JointUPtr joint(new Joint(jointList_.size(), name, size, parent,
            localPosition, eulerAxis0, eulerAngle0));
        AddJointToMapList(joint.get());
        parent->AddChild(move(joint));
    }
}

void Tree::AddJoint(const std::string& name, float size,
                    const std::string& parentName,
                    const Eigen::Vector3f& localPosition,
                    const Eigen::Vector3f& eulerAxis0, float eulerAngle0,
                    const Eigen::Vector3f& eulerAxis1, float eulerAngle1)
{
    if (jointList_.size() == 0) { // this is the root
        root_.reset(new Joint(jointList_.size(), name, size, nullptr,
            localPosition, eulerAxis0, eulerAngle0, eulerAxis1, eulerAngle1));
        AddJointToMapList(root_.get());
    } else { // this is not the root
        auto parent = GetJoint(parentName);
        ASSERT(parent != nullptr, "Parent does not exist.");
        JointUPtr joint(new Joint(jointList_.size(), name, size, parent,
            localPosition, eulerAxis0, eulerAngle0, eulerAxis1, eulerAngle1));
        AddJointToMapList(joint.get());
        parent->AddChild(move(joint));
    }
}

void Tree::AddJoint(const std::string& name, float size,
                    const std::string& parentName,
                    const Eigen::Vector3f& localPosition,
                    const Eigen::Vector3f& eulerAxis0, float eulerAngle0,
                    const Eigen::Vector3f& eulerAxis1, float eulerAngle1,
                    const Eigen::Vector3f& eulerAxis2, float eulerAngle2)
{
    if (jointList_.size() == 0) { // this is the root
        root_.reset(new Joint(jointList_.size(), name, size, nullptr,
            localPosition, eulerAxis0, eulerAngle0, eulerAxis1, eulerAngle1,
            eulerAxis2, eulerAngle2));
        AddJointToMapList(root_.get());
    } else { // this is not the root
        auto parent = GetJoint(parentName);
        ASSERT(parent != nullptr, "Parent does not exist.");
        JointUPtr joint(new Joint(jointList_.size(), name, size, parent,
            localPosition, eulerAxis0, eulerAngle0, eulerAxis1, eulerAngle1,
            eulerAxis2, eulerAngle2));
        AddJointToMapList(joint.get());
        parent->AddChild(move(joint));
    }
}

bool Tree::Contain(const string& name) const
{
    return GetJoint(name) != nullptr;
}

const Joint* Tree::GetRoot() const { return root_.get(); }

Joint* Tree::GetJoint(const string& name) {
    auto constSelf = const_cast<const Tree*>(this);
    return const_cast<Joint*>(constSelf->GetJoint(name));
}

int Tree::GetJointCount() const { return jointList_.size(); }

const Joint* Tree::GetJoint(const std::string& name) const {
    auto item = jointMap_.find(name);
    if (item == jointMap_.end())
        return nullptr;
    else
        return item->second;
}

void Tree::EnableRootTranslation() { isRootMovable_ = true; }

void Tree::DisableRootTranslation() { isRootMovable_ = false; }

bool Tree::IsRootMovable() const { return isRootMovable_; }

void Tree::ComputeGlobalData() {
    for (auto joint = Begin(); joint != End(); ++joint)
        joint->ComputeGlobalData();
}

void Tree::TranslateRoot(float x, float y, float z) {
    root_->AddToLocalPosition(x, y, z);
}

void Tree::AddJointToMapList(Joint* joint) {
    jointMap_[joint->name_] = joint;
    jointList_.push_back(joint);
}

Joint& Dereference(Joint* joint) { return *joint; }

Joint const& DereferenceConst(const Joint* joint) { return *joint; }

JointIterator Tree::Begin() {
    return make_transform_iterator(jointList_.begin(), Dereference);
}

JointIterator Tree::End() {
    return make_transform_iterator(jointList_.end(), Dereference);
}

JointIteratorConst Tree::CBegin() const {
    return make_transform_iterator(jointList_.cbegin(), DereferenceConst);
}

JointIteratorConst Tree::CEnd() const {
    return make_transform_iterator(jointList_.cend(), DereferenceConst);
}