#ifndef TREE_H__
#define TREE_H__

#include <map>
#include <boost/iterator/transform_iterator.hpp>
#include "Common.h"
#include "Joint.h"

// A tree whose nodes are joints.
class Tree
{
private:
    typedef std::unique_ptr<Joint> JointUPtr;
    typedef std::map<std::string, Joint*> JointMap;
    typedef std::vector<Joint*> JointList;

    // We need these to use Boost's transform_iterator to transform an iterator
    // to Joint* to an iterator to Joint.
    typedef Joint& (*DerefFunc)(Joint* joint);
    typedef const Joint& (*DerefCFunc)(const Joint* joint);
    typedef std::vector<Joint*>::iterator VecIter;
    typedef std::vector<Joint*>::const_iterator VecCIter;

public:
    // (Const) Iterator to Joint for internal use.
    typedef boost::transform_iterator<DerefFunc, VecIter> JointIter;
    typedef boost::transform_iterator<DerefCFunc, VecCIter> JointIterConst;

    Tree();
    virtual ~Tree();

    // Adds a hinge joint (1 dof).
    void AddJoint(const std::string& name, float size,
                  const std::string& parentName,
                  const Eigen::Vector3f& localPosition,
                  const Eigen::Vector3f& eulerAxis0, float eulerAngle0);
    // Adds a saddle joint (2 dof).
    void AddJoint(const std::string& name, float size,
                  const std::string& parentName,
                  const Eigen::Vector3f& localPosition,
                  const Eigen::Vector3f& eulerAxis0, float eulerAngle0,
                  const Eigen::Vector3f& eulerAxis1, float eulerAngle1);
    // Adds a ball joint (3 dof).
    void AddJoint(const std::string& name, float size,
                  const std::string& parentName,
                  const Eigen::Vector3f& localPosition,
                  const Eigen::Vector3f& eulerAxis0, float eulerAngle0,
                  const Eigen::Vector3f& eulerAxis1, float eulerAngle1,
                  const Eigen::Vector3f& eulerAxis2, float eulerAngle2);

    bool Contain(const std::string& name) const;

    // Gets the root.
    const Joint* GetRoot() const;
    // Gets the number of joints.
    int GetJointCount() const;
    const Joint* GetJoint(const std::string& name) const;

    // Makes the root movable.
    void EnableRootTranslation();
    // Makes the root stays fixed.
    void DisableRootTranslation();
    // Returns true if the root is movable.
    bool IsRootMovable() const;

    // Compute the joints' global positions and orientations.
    void ComputeGlobalData();

    // Translates the root using v = (dX, dY, dZ).
    void TranslateRoot(float dX, float dY, float dZ);

    // Gets a (const) iterator to the beginning/end of the joint list.
    JointIter Begin();
    JointIter End();
    JointIterConst CBegin() const;
    JointIterConst CEnd() const;

private:
    // Adds the given joint to the map and list.
    void AddJointToMapList(Joint* joint);
    // Gets the joint whose name is name. Returns nullptr if nothing is found.
    Joint* GetJoint(const std::string& name);

    // The root.
    JointUPtr root_;
    // True if the root is movable, false otherwise.
    bool isRootMovable_;
    // We use this to quickly find a joint with a given name.
    JointMap jointMap_;
    // We use this to iterate through the joints in a parent-first order (for
    // all joints n, if never visit n before we visit n's parent).
    JointList jointList_;
};

// (Const) Iterator to Joint for external use.
typedef Tree::JointIter JointIterator;
typedef Tree::JointIterConst JointIteratorConst;

#endif    // TREE_H__